/******************************************************************************
* To the extent possible under law, Sebastian Roith
* <https://www.roith-unterrichtet.de> has waived all copyright and related or
* neighboring rights to this script. This work is published from Germany.
******************************************************************************/

/** Klasse zum Verwalten und Beleben einer Simulation einer optischen Bank */
function OpticalBench(
    offset,             // x-Koordinate des Ursprungs der Skala in lokalen SVG-Einheiten
    nodeBild,           // SVG-Gruppenelement, welches von Schirm eingefangenes Bild darstellt
    nodeFilterElement,  // SVG-feGaussianBlur-Element, welches Unschärfe des eingefangenen Bildes festlegt
    nodeSchirm,         // SVG-Gruppenelement, welches Schirm darstellt
    nodeLinse,          // SVG-Gruppenelement, welches Linse darstellt
    nodeSender          // Optional: SVG-Gruppenelement, welches abzubildenden Gegenstand darstellt
) {
    Object.defineProperty(this, 'offset', { enumerable: true, value: offset });
    Object.defineProperty(this, 'nodeBild', { enumerable: true, value: nodeBild });
    Object.defineProperty(this, 'nodeFilterElement', { enumerable: true, value: nodeFilterElement });
    Object.defineProperty(this, 'nodeSchirm', { enumerable: true, value: nodeSchirm });
    Object.defineProperty(this, 'nodeLinse', { enumerable: true, value: nodeLinse });
    Object.defineProperty(this, 'nodeSender', { enumerable: true, value: nodeSender ? nodeSender : null });

    let svgElement = nodeLinse;
    for (svgElement; svgElement; svgElement = svgElement.parentNode) {
        if (svgElement.nodeName.toUpperCase() == 'SVG')
            break;
    }
    Object.defineProperty(this, 'svgElement', { enumerable: true, value: svgElement });

    this._positions = [undefined, undefined, undefined];
    this._enableDragging();
}

/**
* Optische Geräte mit Maus/Touch verschiebbar machen
* @return undefined
*/
OpticalBench.prototype._enableDragging = function() {
    let moveStartX, moveStartY, movingIndex, movingOffset;
    const handleDragStart = (event) => {
        if (event.touches instanceof Array && event.touches.length != 1)
            return;
        const svgPos = this.svgElement.getBoundingClientRect();
        moveStartX = (event.clientX || event.touches[0].clientX) - svgPos.left;
        moveStartY = (event.clientY || event.touches[0].clientY) - svgPos.top;
        const localX = this.getLocalX(moveStartX);
        const what = this._positions
            .map((pos, index) => ({ pos, index, offset: localX - pos }))
            .filter(({ offset }) => Math.abs(offset) <= 6)
            .reduce((a, b) => (a.pos > b.pos) ? a : b, { pos: -Infinity });
        if (!what || what.index === undefined)
            return;
        movingIndex = what.index;
        movingOffset = what.offset;
        document.addEventListener('mousemove', handleDragging);
        document.addEventListener('touchmove', handleDragging);
        document.addEventListener('mouseup', handleDragEnd);
        document.addEventListener('touchend', handleDragEnd);
    };

    const handleDragging = (event) => {
        const svgPos = this.svgElement.getBoundingClientRect();
        let moveEndX = (event.clientX || event.touches[0].clientX) - svgPos.left;
        let moveEndY = (event.clientY || event.touches[0].clientY) - svgPos.top;
        const deltaX = moveEndX - moveStartX;
        const deltaY = moveEndY - moveStartY;
        if (Math.abs(deltaY / deltaX) > 0.1 && Math.abs(deltaY) > 30)
            return;
        this._updatePos(this.getLocalX(moveEndX) - movingOffset, movingIndex);
    };

    function handleDragEnd() {
        document.removeEventListener('mousemove', handleDragging);
        document.removeEventListener('touchmove', handleDragging);
        document.removeEventListener('mouseup', handleDragEnd);
        document.removeEventListener('touchend', handleDragEnd);
    }

    this.svgElement.addEventListener('mousedown', handleDragStart);
    this.svgElement.addEventListener('touchstart', handleDragStart);
}

// Brennweite in Zentimetern
OpticalBench.prototype.brennweite = 10;

// Linsenradius in Zentimetern
OpticalBench.prototype.linsenradius = 10;

// Abstahlwinkel des Senders in Grad
OpticalBench.prototype.abstrahlwinkel = 15;

// Korrekturfaktor für Bildgröße
OpticalBench.prototype.korrekturBildgroesse = 1;

// Minimal erlaubter x-Wert (ausgenommen undefined und ggf. -Infinity oder +Infinity)
OpticalBench.prototype.minPos = 0;

// Maximal erlaubter x-Wert (ausgenommen undefined und ggf. -Infinity oder +Infinity)
OpticalBench.prototype.maxPos = 100;

// Minimaler Abstand zwischen zwei optischen Geräten
OpticalBench.prototype.minDistance = 3;

// Funktion, die aufgerufen wird, wenn sich die Position eines optischen Gerätes ändert
OpticalBench.prototype.onchange = function(index, newValue) {}

// Position von Gegenstand, Linse und Schirm
OpticalBench.prototype._positions = [undefined, undefined, undefined];


// Neue Position eines optischen Gerätes ggf. angepasst übernehmen
OpticalBench.prototype._updatePos = function(
    newPos,     // Neue Position in Zentimetern (oder undefined bzw. bei Sender auch +Infinity oder -Infinity)
    type        // Art des optischen Gerätes: 0 → Sender, 1 → Linse, 2 → Schirm
) {
    let pos = newPos;
    if (Number.isFinite(pos))
        pos = Math.min(this.maxPos, Math.max(this.minPos, Math.round(pos)));
    else
        pos = (!type && Math.abs(pos) == Infinity) ? pos : undefined;
    if (Number.isFinite(pos)) {
        let otherPos = this._positions.slice();
        otherPos.splice(type, 1);
        otherPos = otherPos.filter(v => Number.isFinite(v));
        for (let offset = 0; offset < 100; ++offset) {
            if (pos + offset <= this.maxPos && otherPos.reduce((prev, other) => prev && Math.abs(pos + offset - other) >= this.minDistance, true)) {
                pos += offset;
                break;
            }
            if (pos - offset >= this.minPos && otherPos.reduce((prev, other) => prev && Math.abs(pos - offset - other) >= this.minDistance, true)) {
                pos -= offset;
                break;
            }
        }
    }
    if (pos == this._positions[type])
        return;
    this._positions[type] = pos;
    this.update();
    if (typeof this.onchange == 'function')
        this.onchange.call(this, type, pos);
}

// Position des Schirms
Object.defineProperty(OpticalBench.prototype, 'posSchirm', {
    enumerable: true,
    get() { return this._positions[2]; },
    set(pos) { this._updatePos(pos, 2); }
});

// Position der Linse
Object.defineProperty(OpticalBench.prototype, 'posLinse', {
    enumerable: true,
    get() { return this._positions[1]; },
    set(pos) { this._updatePos(pos, 1); }
});

// Position des Schirms
Object.defineProperty(OpticalBench.prototype, 'posSender', {
    enumerable: true,
    get() { return this._positions[0]; },
    set(pos) {
        if (this.nodeSender)
            this._updatePos(pos, 0);
        else
            this._positions[0] = pos;
    }
});

// Eingestellte Gegenstandsweite
Object.defineProperty(OpticalBench.prototype, 'gegenstandsweite', {
    enumerable: true,
    get() { return this.posLinse - this.posSender; }
});

// Eingestellte Schirmweite
Object.defineProperty(OpticalBench.prototype, 'schirmweite', {
    enumerable: true,
    get() { return this.posSchirm - this.posLinse; }
});

// Eigentliche Bildweite
Object.defineProperty(OpticalBench.prototype, 'bildweite', {
    enumerable: true,
    get() {
        const gegenstandsweite = this.gegenstandsweite;
        return gegenstandsweite * this.brennweite / (Math.abs(gegenstandsweite) - this.brennweite);
    }
});

/**
* Verschiebung in x-Richtung festlegen
* @return undefined
*/
OpticalBench.setX = function(
    node,   // SVG-Element, dessen Position geändert werden soll
    x       // Gewünschte Verschiebung vom Start (oder nicht-endliche Zahl für ausblenden)
) {
    if (node)
        node.style.transform = 'translateX(' + (Number.isFinite(x) ? x : '-10000') + 'px)';
}

/**
* Position und Schirminhalt anpassen
* @return undefined
*/
OpticalBench.prototype.update = function() {
    OpticalBench.setX(this.nodeSender, this.posSender);
    OpticalBench.setX(this.nodeLinse,  this.posLinse);
    OpticalBench.setX(this.nodeSchirm, this.posSchirm);
    this._showOnScreen();
    this._repairOrder();
}

/**
* Anzeige auf Schirm entsprechend berechneter Größe, Schärfe und Richtung tatsächlich anpassen
* @return undefined
*/
OpticalBench.prototype._display = function(
    radius,         // Größe der Anzeige (Radius bei einfarbigen Kreis)
    scharfradius,   // Bis zu welchem Radius wäre ein einfarbiger Kreis scharf?
    gespiegelt      // Ist das Bild gespiegelt?
) {
    const bildgroesse = radius * this.korrekturBildgroesse;
    this.nodeBild.setAttribute('transform', 'scale(' + (gespiegelt ? -1 : +1) * (Number.isFinite(bildgroesse) ? bildgroesse : 999999) + ')');
    const opacity = Math.min(1, Math.max(0, Math.sqrt(2 / ( radius + scharfradius))));
    this.nodeBild.style.opacity = opacity;

    const stdDeviation = (radius - scharfradius) / radius / 2;
    this.nodeFilterElement.setAttribute('stdDeviation', Number.isFinite(stdDeviation) ? stdDeviation : 999999);
}

/**
* Was auf dem Schirm angezeigt wird neu berechnen und Rechenergebnisse anwenden
* @return undefined
*/
OpticalBench.prototype._showOnScreen = function() {
    if (!Number.isFinite(this.posSchirm))
        return;
    if (!Number.isFinite(this.posSender)) {
        this._display(999999, 0, false);
        return;
    }
    const gegenstandsweite = this.gegenstandsweite;
    const schirmweite = this.schirmweite;
    if (!Number.isFinite(this.posLinse) || Math.sign(gegenstandsweite) != Math.sign(schirmweite)) {
        const abstand = Math.abs(this.posSender - this.posSchirm);
        const radius = abstand * Math.tan(this.abstrahlwinkel * Math.PI / 180) + 1;
        this._display(Math.min(radius, 999999), 0, false);
        return;
    }
    const schirmpos = schirmweite * Math.sign(gegenstandsweite);
    const bildweite = Math.abs(this.bildweite);
    const bildgroesse = Math.abs(1 * bildweite / gegenstandsweite);
    const radius1 = Math.abs(this.linsenradius + schirmpos * (bildgroesse - this.linsenradius) / bildweite);
    const radius2 = Math.abs(schirmpos * (bildgroesse + this.linsenradius) / bildweite - this.linsenradius);
    this._display(Math.max(radius1, radius2), Math.min(radius1, radius2), true);
}

/**
* Stapelreihenfolge der optischen Geräte in der Darstellung ggf. an Positionen anpassen
* @return undefined
*/
OpticalBench.prototype._repairOrder = function() {
    function getChildIndex(node) {
        const parent = node.parentNode;
        return Array.from(parent.childNodes).indexOf(node);
    }

    const infos = this._positions
        .map((pos, index) => ({ pos, node: [this.nodeSender, this.nodeLinse, this.nodeSchirm][index] }))
        .filter(obj => obj.node && Number.isFinite(obj.pos));
    if (infos.length < 2)
        return;
    infos.sort((a, b) => a.pos - b.pos);
    infos.reduce((a, b) => {
        if (getChildIndex(a.node) > getChildIndex(b.node))
            b.node.parentNode.insertBefore(a.node, b.node);
        return b;
    });
}

/**
* Eine x-Koordinate in Seitenpixeln relative zum SVG-Element umrechnen
* @return number    Position in lokalen SVG-Einheiten
*/
OpticalBench.prototype.getLocalX = function(
    x       // x-Koordinate in Seitenpixeln
) {
    return x * parseFloat(this.svgElement.getAttribute('width')) / this.svgElement.clientWidth - 4;
}

/**
* Ein Eingabefeld künftig zur Anzeige und zum Bearbeiten einer Position verwenden
* @return undefined
*/
OpticalBench.prototype.registerPositionInput = function(
    postype,    // 0 für Sender, 1 für Linse oder 2 für Schirm
    nodeInput   // Eingabefeld
) {
    const showPos = () => nodeInput.value = Number.isFinite(this._positions[postype]) ? this._positions[postype] : '';
    showPos();
    let focused = false;
    nodeInput.addEventListener('input', () => this._updatePos(Number.parseInt(nodeInput.value), postype));
    nodeInput.addEventListener('focus', () => focused = true);
    nodeInput.addEventListener('blur', () => { focused = false; showPos(); });
    const fktChange = (typeof this.onchange == 'function') ? this.onchange : function(){};
    this.onchange = (type, newValue) => {
        if (!focused && type == postype)
            showPos();
        fktChange.call(this, type, newValue);
    };
}
